/*
  * Originates in the work for a seminar in the winter semester 2000/2001
  * at the University of Essen (http://www.uni-essen.de), department:
  * "MIS and software development" (http://nestroy.wi-inf.uni-essen.de).
  *
  * This DLL registers the function "BSF" with Rexx;
  * it allows Java to call into Rexx using the fine RexxSAA-interface
  * conventions and JNI ("Java Native Interface").
  *
  * version: 1.1.2, 2001-06-02, ---rgf, runs with JNI 1 (does *not* need JRE 1.2 or higher anymore)
  *
  * author:  (c) 2001 Rony G. Flatscher, University of Essen, WU  (pronounced: "vey-u") University Vienna
  *                  (Peter Kalender (Oct 2000 to Feb 2001), University of Essen)
  *
  *  LICENSE:
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
  *  License as published by the Free Software Foundation; either
  *  version 2.1 of the License, or (at your option) any later version.
  *
  *  This library is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  *  Lesser General Public License for more details.
  *
  *  You should have received a copy of the GNU Lesser General Public
  *  License along with this library; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *
  *  The following URL may be available for further information:
  *
  *     <a href="http://www.gnu.org/copyleft/lesser.html">
  *              http://www.gnu.org/copyleft/lesser.html
  *     </a>

*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#define INCL_RXSHV
#define INCL_RXFUNC

#ifdef USE_REXXTRANS
# include <rexxtrans.h>
#elif defined USE_OREXX
# include <rexx.h>
#else
# include <rexxsaa.h>
#endif

#ifdef _MSC_VER
# pragma warning(disable:4100)
#endif



#ifdef __cplusplus
extern "C" {            // make sure no C++ name mangling comes into the way
#endif


#define RFH_BSF         "BSF"           // name of BSF function handler
#define DLLNAME         "BSF4Rexx"      // name of DLL and name for "PARSE SOURCE"

        // Java support
#include <jni.h>                        // "Java Native Interface" (JNI)
        // generated by Java to support the JNI
        // see comments at the end of "RexxAndJava.java"
#include "com_ibm_bsf_engines_rexx_RexxAndJava.h"

#define JAVA_CALL_BSF   "javaCallBSF"   // name of Java method to carry out the BSF-call



JNIEnv *globEnv;        // used by BSF() to call back to Java
jobject globObj;        // used by BSF() to call back to Java
jmethodID mid;          // used by BSF() to call back to Java



        // inline function for taking care of the different ways of freeing Rexx-allocated memory
inline void befreieSpeicher (void * memory)
{

#ifdef USE_REGINA
//    printf(">>> befreieSpeicher(): 'USE_REGINA' defined, RexxFreeMemory()\n");
    RexxFreeMemory( memory );

#elif defined WIN32
//    printf(">>> befreieSpeicher(): 'WIN32' defined, GlobalFree()\n");
    GlobalFree( memory );

#elif defined OS2
//    printf(">>> befreieSpeicher(): 'OS2' defined, DosFreeMem()\n");
    DosFreeMem( memory );

#else
//    printf(">>> befreieSpeicher():  using free()\n");
    free( memory );
#endif

}



/*
 * External Rexx function. This function allows Rexx programs to call into Java.
 */
APIRET APIENTRY
BSF  (  PSZ name,
        ULONG argc,
        PRXSTRING argv,
        PSZ queuename,
        PRXSTRING retstr)
     {
        //convert arguments array to Java format
        jobjectArray arr = (*globEnv).NewObjectArray (
                                      argc,
                                    (*globEnv).FindClass("java/lang/String"),
                                    (*globEnv).NewStringUTF("")
                                                     );

        // RXSTRINGs passed from Rexx have an additional \0 appended according Object Rexx 2.1 docs
        for(ULONG i=0; i<argc; i++)
        {
            if (argv[i].strptr==NULL)   // NULL (argument missing) ?
            {
               (*globEnv).SetObjectArrayElement(arr, i, NULL);
            }
            else                        // argument as a string in hand
            {
               (*globEnv).SetObjectArrayElement(arr, i,
                                               (*globEnv).NewStringUTF(argv[i].strptr));
            }
        }

        // call Java method and create Rexx return value
        jstring ReturnValue = (jstring)(*globEnv).CallObjectMethod(globObj, mid, arr);


        if (ReturnValue!=NULL)          // value returned from Java?
        {
                const char *str = globEnv->GetStringUTFChars(ReturnValue, 0);
                int         len = strlen(str);

                // o.k. not enough space available, free Rexx-allocated memory, re-allocate new chunk
                if (len>255)
                {
                   befreieSpeicher ( retstr->strptr );
                   retstr->strptr = new char[len+1];
                }

                strcpy(retstr->strptr, str);
                retstr->strlength = len;
                globEnv->ReleaseStringUTFChars(ReturnValue, str);
        }
        else    // o.k. Java returned null
        {
            // rgf, 2001-05
            if (retstr->strptr != NULL)
            {
                 befreieSpeicher ( retstr->strptr );
            }

            retstr->strptr    = NULL;   // set string pointer to NULL (as if no value was returned)
            retstr->strlength = 0;      // indicate no chars available
        }

                // pending exception ?
        if ( globEnv->ExceptionOccurred() != NULL ) // this works on JRE 1.1.8 ! ---rgf, 2001-05-25
        {
             globEnv->ExceptionDescribe();      // print the stack
             globEnv->ExceptionClear();         // clear exception in the JVM
             return 40l;        // interrupt Rexx with "Incorrect call to routine!"
        }
        return 0l;
     }



/*
 * This will be called from Java at initialization time and makes sure
 * that the external Rexx function named according to RFH_BSF will
 * registered. This way, Rexx programs being invoked from the Java side
 * have RFH_BSF available already (no need to register this function by
 * called Rexx program). Changed to "jint" on 2001-05-25.
 */
JNIEXPORT jint JNICALL
Java_com_ibm_bsf_engines_rexx_RexxAndJava_jniRegisterBSF (JNIEnv *env, jobject obj)
     {
         return (jint) RexxRegisterFunctionDll(RFH_BSF , DLLNAME, RFH_BSF );
     }



/*
 * This will be called from Java to invoke the Rexx interpreter. "source" contains
 * the Rexx program, "array" the arguments meant for the Rexx program.
 */
JNIEXPORT jstring JNICALL Java_com_ibm_bsf_engines_rexx_RexxAndJava_jniRexxStart
        ( JNIEnv *env,
          jobject         obj,
          jstring         source,
          jobjectArray    array)
     {
         // set global variables for callbacks to Java (via BSF())
         globEnv = env;
         globObj = obj;

                // get Java callback method via the Java object's class, looking up a method
                // named according to JAVA_CALL_BSF, signature: one argument (a String array) and
                // returning a String
         mid = (*globEnv).GetMethodID( (*globEnv).GetObjectClass(globObj),
                                       JAVA_CALL_BSF,
                                       "([Ljava/lang/String;)Ljava/lang/String;" );

         // make a copy of the Java string containing the Rexx script to be executed
         const char      *str = env->GetStringUTFChars(source, 0);
         char      *sourceStr = new char[strlen(str)+1];
         strcpy(sourceStr, str);
         env->ReleaseStringUTFChars(source, str);

         // create argument array for Rexx
         PRXSTRING argList = NULL;       // pointer to array of RXSTRINGS
         jsize   arrNum = 0;
         long    numArgs= 0l;

         if (array != NULL)              // an Object array available?
         {
             arrNum    =  env->GetArrayLength(array);
             numArgs   = (long) arrNum;
         }

         if (arrNum>0)
         {
            argList = new RXSTRING [ arrNum ];
            for (int i = 0; i < arrNum; ++i)
            {
               if ((env->GetObjectArrayElement(array, i))==NULL)/* no argument*/
               {
                  argList[i].strptr   =NULL;
                  argList[i].strlength=0;
               }
               else        /* a string argument in hand, give it to Rexx*/
               {
                  jstring        javaStr = (jstring) env->GetObjectArrayElement(array, i);
                  const char     *strArg = env->GetStringUTFChars(javaStr, 0);
                  int                len = strlen(strArg);
                  char        *tmpString = new char[len+1];
                  strcpy(tmpString, strArg);
                  env->ReleaseStringUTFChars(javaStr, strArg);

                  MAKERXSTRING( argList[i], tmpString, len);/* save arg for Rexx*/
               }
            }
         }

                // Rexx program to execute
         RXSTRING        RXsourceStr[2];
         MAKERXSTRING   (RXsourceStr[0], sourceStr, strlen(sourceStr));
         PRXSTRING       Instore = &RXsourceStr[0];
         MAKERXSTRING   (RXsourceStr[1], NULL, 0);

         SHORT           ReturnCode;
         LONG            ReturnValue;

                // let Rexx allocate buffer, if needed for a result
         RXSTRING        Result;
         MAKERXSTRING   (Result, NULL, 0);

         PSZ             ProgramName = DLLNAME ;

         //start Rexx
         ReturnValue = RexxStart(
                         arrNum,         // LONG         ArgCount
                         argList,        // PRXSTRING    ArgList
                         ProgramName,    // PSZ          ProgramName
                         Instore,        // PRXSTRING    Instore
                         NULL,           // PSZ          EnvName
                         RXSUBROUTINE,   // LONG         CallType
                         NULL,           // PRXSYSEXIT   Exits
                         &ReturnCode,    // PUSHORT      ReturnCode
                         &Result         // PRXSTRING    Result
                         );


                // freeing memory for Rexx program and tokenized image
        befreieSpeicher ( RXsourceStr[0].strptr );// free Rexx program
        befreieSpeicher ( RXsourceStr[1].strptr );// free tokenized image of it


        if (arrNum>0)   // free memory needed for Rexx arguments
        {
// printf("Java_..._jniRexxStart(): Freeing memory for Rexx arguments...\n");

           for (int i = 0; i < arrNum; ++i)
           {
              befreieSpeicher ( argList[i].strptr );
           }
           delete [] argList;
        }


        if (ReturnValue != 0l)     // error occurred, return null
        {
// printf("Java_..._jniRexxStart(): error occurred, freeing Rexx result-memory at #1 ...\n");

           befreieSpeicher ( Result.strptr );     // free memory allocated from Rexx
           return NULL;
        }

         // --- get result and return it to Java
        jstring jResult=NULL;           // define Java String to return

        if ( Result.strptr != NULL )    // a result was returned from Rexx
        {
             if (Result.strlength==0)   // empty string?
             {
                 jResult=env->NewStringUTF("");
             }
             else                       // valid non-empty string
             {
                 jResult=env->NewStringUTF( Result.strptr );
             }

// printf("Java_..._jniRexxStart(): freeing Rexx result-memory at #2 ...\n");

             befreieSpeicher ( Result.strptr ); // free memory where Rexx placed its result into
        }

        return jResult;         // return a Java String
     }



#ifdef __cplusplus
}               // closing bracket
#endif


